"use strict";
var ts = require('typescript');
var path_1 = require('path');
var source_map_1 = require('source-map');
function mahaloTranspiler(moduleName, shouldDiagnose) {
    if (shouldDiagnose === void 0) { shouldDiagnose = false; }
    var fileName = resolveModuleName(moduleName);
    var program = ts.createProgram([fileName], { moduleResolution: ts.ModuleResolutionKind.NodeJs });
    var checker = program.getTypeChecker();
    var sourceFile = program.getSourceFile(fileName);
    var text = sourceFile.text;
    var edits = [];
    var identifiers = [];
    var assignIdentifier = 'assign';
    var promiseIdentifier = 'Promise';
    var shouldImportAssign = false;
    var shouldImportPromise = true;
    var assignmentDepth = 0;
    var mahaloMap;
    shouldDiagnose && [
        // 'getDeclarationDiagnostics',
        'getGlobalDiagnostics',
        'getOptionsDiagnostics',
        'getSemanticDiagnostics',
        'getSyntacticDiagnostics'
    ].forEach(function (method) { return program[method]().forEach(function (diagnostic) {
        if (diagnostic.category === ts.DiagnosticCategory.Error) {
            throw Error(diagnostic.messageText.toString());
        }
    }); });
    if (!/\/(mahalo|core-js)\//.test(fileName)) {
        findIdentifiers(sourceFile);
        while (identifiers.indexOf(assignIdentifier) > -1) {
            assignIdentifier = '_' + assignIdentifier;
        }
        while (identifiers.indexOf(promiseIdentifier) > -1) {
            promiseIdentifier = '_' + promiseIdentifier;
        }
        findEdits(sourceFile);
        importModules();
        var edited = applyEdits(path_1.basename(fileName), text, edits);
        text = edited.text;
        mahaloMap = edited.map;
    }
    var result = ts.transpileModule(text, {
        fileName: fileName,
        compilerOptions: {
            target: ts.ScriptTarget.ES5,
            sourceMap: true
        }
    });
    return {
        fileName: fileName,
        text: result.outputText.replace(/(\r\n|\r|\n)\/\/# sourceMappingURL=.*\.js\.map$/, ''),
        map: mergeSourceMaps(mahaloMap, JSON.parse(result.sourceMapText))
    };
    //////////
    function findIdentifiers(node) {
        switch (node.kind) {
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.SourceFile:
                Object.keys(node['locals']).forEach(function (identifier) { return identifiers.push(identifier); });
        }
        ts.forEachChild(node, findIdentifiers);
    }
    function findEdits(node) {
        switch (node.kind) {
            case ts.SyntaxKind.BinaryExpression:
                switch (node.operatorToken.kind) {
                    case ts.SyntaxKind.FirstAssignment:
                    case ts.SyntaxKind.PlusEqualsToken:
                    case ts.SyntaxKind.MinusEqualsToken:
                    case ts.SyntaxKind.AsteriskEqualsToken:
                    case ts.SyntaxKind.SlashEqualsToken:
                    case ts.SyntaxKind.PercentEqualsToken:
                    case ts.SyntaxKind.AsteriskAsteriskEqualsToken:
                    case ts.SyntaxKind.BarEqualsToken:
                    case ts.SyntaxKind.AmpersandEqualsToken:
                    case ts.SyntaxKind.GreaterThanGreaterThanEqualsToken:
                    case ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken:
                    case ts.SyntaxKind.LessThanLessThanEqualsToken:
                    case ts.SyntaxKind.CaretEqualsToken:
                        createAssignmentEdit(node);
                        assignmentDepth++;
                        ts.forEachChild(node, findEdits);
                        assignmentDepth--;
                        return;
                }
                break;
            case ts.SyntaxKind.PrefixUnaryExpression:
            case ts.SyntaxKind.PostfixUnaryExpression:
                createUpdateEdit(node);
                break;
            case ts.SyntaxKind.DeleteExpression:
                createDeleteEdit(node);
                break;
            case ts.SyntaxKind.ClassDeclaration:
                createRouteEdit(node);
                break;
        }
        ts.forEachChild(node, findEdits);
    }
    function createAssignmentEdit(node) {
        var change = {
            start: node.getStart(),
            length: node.getWidth(),
            newText: '',
            depth: assignmentDepth
        };
        var operatorToken = node.operatorToken;
        var left = node.left;
        if (left.kind === ts.SyntaxKind.PropertyAccessExpression) {
            var object = left.expression.getText();
            var key = left.name.getText();
            var value = node.right.getText();
            if (operatorToken.kind !== ts.SyntaxKind.FirstAssignment) {
                var operator = operatorToken.getText().substr(0, operatorToken.getWidth() - 1);
                value = left.getText() + " " + operator + " " + value;
            }
            change.newText = assignIdentifier + "(" + object + ", " + JSON.stringify(key) + ", " + value + ")";
        }
        else {
            change.newText = assignIdentifier + "(" + node.getText() + ")";
        }
        edits.push(change);
        shouldImportAssign = true;
    }
    function createUpdateEdit(node) {
        if ([ts.SyntaxKind.PlusPlusToken, ts.SyntaxKind.MinusMinusToken].indexOf(node.operator) === -1) {
            return;
        }
        var operator = node.operator === ts.SyntaxKind.PlusPlusToken ? '+' : '-';
        var change = {
            start: node.getStart(),
            length: node.getWidth()
        };
        var operand = node.operand;
        if (operand.kind === ts.SyntaxKind.PropertyAccessExpression) {
            var object = operand.expression.getText();
            var key = operand.name.getText();
            var value = operand.getText() + " " + operator + " 1";
            change.newText = assignIdentifier + "(" + object + ", " + JSON.stringify(key) + ", " + value + ")";
            if (node.kind === ts.SyntaxKind.PostfixUnaryExpression) {
                change.newText = "(" + change.newText + " " + (operator === '+' ? '-' : '+') + " 1)";
            }
        }
        else {
            change.newText = assignIdentifier + "(" + node.getText() + ")";
        }
        edits.push(change);
        shouldImportAssign = true;
    }
    function createDeleteEdit(node) {
        var expression = node.expression;
        if (node.expression.kind !== ts.SyntaxKind.PropertyAccessExpression) {
            return;
        }
        var object = expression.expression.getText();
        var key = expression.name.getText();
        edits.push({
            start: node.getStart(),
            length: node.getWidth(),
            newText: assignIdentifier + "(" + object + ", " + JSON.stringify(key) + ")"
        });
        shouldImportAssign = true;
    }
    function createRouteEdit(node) {
        if (!extendsRoute(node)) {
            return;
        }
        for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
            var member = _a[_i];
            if (member.name.getText() === 'view') {
                var initializer = member['initializer'];
                if (initializer && initializer.kind === ts.SyntaxKind.StringLiteral) {
                    var view = initializer.text;
                    edits.push({
                        start: initializer.getStart(),
                        length: initializer.getWidth(),
                        newText: "() => {\n                            return new " + promiseIdentifier + "(resolve => {\n                                require.ensure(" + JSON.stringify(view) + ", require => {\n                                    resolve(require(" + JSON.stringify(view) + "));\n                                });\n                            });\n                        }"
                    });
                    shouldImportPromise = true;
                }
                return;
            }
        }
    }
    function extendsRoute(node) {
        if (node.heritageClauses) {
            for (var _i = 0, _a = node.heritageClauses; _i < _a.length; _i++) {
                var clause = _a[_i];
                if (clause.token === ts.SyntaxKind.ExtendsKeyword) {
                    var symbolForType = checker.getTypeAtLocation(clause.types[0].expression).symbol;
                    if (symbolForType.name === 'default' && symbolForType['parent']) {
                        if (/\/node_modules\/mahalo\/components\/route"$/.test(symbolForType['parent'].name)) {
                            return true;
                        }
                        if (symbolForType.valueDeclaration.kind === ts.SyntaxKind.ClassDeclaration) {
                            return extendsRoute(symbolForType.valueDeclaration);
                        }
                    }
                    break;
                }
            }
        }
        return false;
    }
    function importModules() {
        shouldImportPromise && edits.unshift({
            start: 0,
            length: 0,
            newText: "import * as " + promiseIdentifier + " from 'core-js/library/es6/promise';\n"
        });
        shouldImportAssign && edits.unshift({
            start: 0,
            length: 0,
            newText: "import {assign as " + assignIdentifier + "} from 'mahalo';\n"
        });
    }
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = mahaloTranspiler;
//////////
function resolveModuleName(moduleName) {
    var host = ts.createCompilerHost({ allowJs: true });
    var file = ts.nodeModuleNameResolver(moduleName, '', { allowJs: true }, host);
    if (!file) {
        throw Error('Cannot find module ' + moduleName);
    }
    return path_1.resolve(file.resolvedModule.resolvedFileName).replace(/\\/g, '/');
}
function applyEdits(fileName, text, edits) {
    var result = text;
    var mapGenerator = new source_map_1.SourceMapGenerator();
    var change = edits[0];
    var i = 0;
    var delta = 0;
    while (change) {
        var depth = change.depth || 0;
        var start = change.start + delta - depth;
        var head = result.substr(0, start);
        var tail = result.substr(start + change.length);
        delta += change.newText.length - change.length;
        result = head + change.newText + tail;
        mapGenerator.addMapping({
            generated: getLineAndColumn(result, start),
            original: getLineAndColumn(text, change.start),
            source: fileName
        });
        change = edits[++i];
    }
    return {
        text: result,
        map: JSON.parse(mapGenerator.toString())
    };
}
/**
 * Merges the source maps provided by each of the two steps.
 */
function mergeSourceMaps(mahaloMap, typescriptMap) {
    if (!mahaloMap) {
        return typescriptMap;
    }
    var mahaloMapConsumer = new source_map_1.SourceMapConsumer(mahaloMap);
    var typescriptMapConsumer = new source_map_1.SourceMapConsumer(typescriptMap);
    var mergedMapGenerator = new source_map_1.SourceMapGenerator();
    typescriptMapConsumer.eachMapping(function (mapping) {
        var originalPosition = mahaloMapConsumer.originalPositionFor({
            line: mapping.originalLine,
            column: mapping.originalColumn
        });
        if (!originalPosition.source) {
            return;
        }
        mergedMapGenerator.addMapping({
            original: {
                line: originalPosition.line,
                column: originalPosition.column
            },
            generated: {
                line: mapping.generatedLine,
                column: mapping.generatedColumn
            },
            source: mapping.source,
            name: mapping.name
        });
    });
    var mergedMap = JSON.parse(mergedMapGenerator.toString());
    mergedMap.sources = mahaloMap.sources;
    return mergedMap;
}
function getLineAndColumn(text, i) {
    var line = 1, column = 0;
    while (i--) {
        text[i] === '\n' && line++;
        line === 1 && column++;
    }
    return { line: line, column: column };
}
